import datetime as dt
import functools
import glob
import logging
import os
import shutil
import time

import numpy as np

from autofit import conf
from autofit.mapper import model_mapper as mm
from autofit.optimize.non_linear.paths import Paths, convert_paths

logging.basicConfig()
logger = logging.getLogger(__name__)  # TODO: Logging issue


class NonLinearOptimizer(object):
    @convert_paths
    def __init__(self, paths):
        """Abstract base class for non-linear optimizers.

        This class sets up the file structure for the non-linear optimizer nlo, which are standardized across \
        all non-linear optimizers.

        Parameters
        ------------

        """
        log_file = conf.instance.general.get("output", "log_file", str).replace(" ", "")
        self.paths = paths

        if not len(log_file) == 0:
            log_path = "{}{}".format(self.paths.phase_output_path, log_file)
            logger.handlers = [logging.FileHandler(log_path)]
            logger.propagate = False
            # noinspection PyProtectedMember
            logger.level = logging._nameToLevel[
                conf.instance.general.get("output", "log_level", str)
                .replace(" ", "")
                .upper()
            ]

        self.restore()

    def config(self, attribute_name, attribute_type=str):
        """
        Get a config field from this optimizer's section in non_linear.ini by a key and value type.

        Parameters
        ----------
        attribute_name: str
            The analysis_path of the field
        attribute_type: type
            The type of the value

        Returns
        -------
        attribute
            An attribute for the key with the specified type.
        """
        return conf.instance.non_linear.get(
            self.__class__.__name__, attribute_name, attribute_type
        )

    def __eq__(self, other):
        return isinstance(other, NonLinearOptimizer) and self.__dict__ == other.__dict__

    def backup(self):
        """
        Copy files from the sym-linked optimizer folder to the backup folder in the workspace.
        """

        try:
            shutil.rmtree(self.paths.backup_path)
        except FileNotFoundError:
            pass

        try:
            shutil.copytree(self.paths.sym_path, self.paths.backup_path)
        except shutil.Error as e:
            logger.exception(e)

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.restore()

    def restore(self):
        """
        Copy files from the backup folder to the sym-linked optimizer folder.
        """
        if os.path.exists(self.paths.backup_path):
            for file in glob.glob(self.paths.backup_path + "/*"):
                shutil.copy(file, self.paths.path)

    def fit(self, analysis, model):
        raise NotImplementedError(
            "Fitness function must be overridden by non linear optimizers"
        )

    class Fitness:
        def __init__(self, nlo, analysis):

            self.nlo = nlo
            self.result = None
            self.max_likelihood = -np.inf
            self.analysis = analysis

            log_interval = conf.instance.general.get("output", "log_interval", int)
            backup_interval = conf.instance.general.get(
                "output", "backup_interval", int
            )
            visualize_interval = conf.instance.visualize.get(
                "figures", "visualize_interval", int
            )

            self.should_log = IntervalCounter(log_interval)
            self.should_backup = IntervalCounter(backup_interval)
            self.should_visualize = IntervalCounter(visualize_interval)

        def fit_instance(self, instance):
            likelihood = self.analysis.fit(instance)

            if likelihood > self.max_likelihood:

                self.max_likelihood = likelihood
                self.result = Result(instance, likelihood)

                if self.should_visualize():
                    self.analysis.visualize(instance, during_analysis=True)

                if self.should_backup():
                    self.nlo.backup()

            return likelihood

    def copy_with_name_extension(self, extension, remove_phase_tag=False):
        name = "{}/{}".format(self.paths.phase_name, extension)

        if remove_phase_tag:
            phase_tag = ""
        else:
            phase_tag = self.paths.phase_tag

        new_instance = self.__class__(
            Paths(
                phase_name=name,
                phase_folders=self.paths.phase_folders,
                phase_tag=phase_tag,
            )
        )

        return new_instance


class Analysis(object):
    def fit(self, instance):
        raise NotImplementedError()

    def visualize(self, instance, during_analysis):
        raise NotImplementedError()


class Result(object):
    """
    @DynamicAttrs
    """

    def __init__(
        self, instance, figure_of_merit, previous_model=None, gaussian_tuples=None
    ):
        """
        The result of an optimization.

        Parameters
        ----------
        instance: autofit.mapper.model.ModelInstance
            An instance object comprising the class instances that gave the optimal fit
        figure_of_merit: float
            A value indicating the figure of merit given by the optimal fit
        previous_model: mm.ModelMapper
            The model mapper from the stage that produced this result
        """
        self.instance = instance
        self.figure_of_merit = figure_of_merit
        self.previous_model = previous_model
        self.gaussian_tuples = gaussian_tuples
        self.__model = None

    @property
    def model(self):
        if self.__model is None:
            self.__model = self.previous_model.mapper_from_gaussian_tuples(
                self.gaussian_tuples
            )
        return self.__model

    @model.setter
    def model(self, model):
        self.__model = model

    def __str__(self):
        return "Analysis Result:\n{}".format(
            "\n".join(
                ["{}: {}".format(key, value) for key, value in self.__dict__.items()]
            )
        )

    def model_absolute(self, a: float) -> mm.ModelMapper:
        """
        Parameters
        ----------
        a
            The absolute width of gaussian priors

        Returns
        -------
        A model mapper created by taking results from this phase and creating priors with the defined absolute
        width.
        """
        return self.previous_model.mapper_from_gaussian_tuples(
            self.gaussian_tuples, a=a
        )

    def model_relative(self, r: float) -> mm.ModelMapper:
        """
        Parameters
        ----------
        r
            The relative width of gaussian priors

        Returns
        -------
        A model mapper created by taking results from this phase and creating priors with the defined relative
        width.
        """
        return self.previous_model.mapper_from_gaussian_tuples(
            self.gaussian_tuples, r=r
        )


class IntervalCounter(object):
    def __init__(self, interval):
        self.count = 0
        self.interval = interval

    def __call__(self):
        if self.interval == -1:
            return False
        self.count += 1
        return self.count % self.interval == 0


def persistent_timer(func):
    """
    Times the execution of a function. If the process is stopped and restarted then timing is continued using
    saved files.

    Parameters
    ----------
    func
        Some function to be timed

    Returns
    -------
    timed_function
        The same function with a timer attached.
    """

    @functools.wraps(func)
    def timed_function(optimizer_instance, *args, **kwargs):
        start_time_path = "{}/.start_time".format(
            optimizer_instance.paths.phase_output_path
        )
        try:
            with open(start_time_path) as f:
                start = float(f.read())
        except FileNotFoundError:
            start = time.time()
            with open(start_time_path, "w+") as f:
                f.write(str(start))

        result = func(optimizer_instance, *args, **kwargs)

        execution_time = str(dt.timedelta(seconds=time.time() - start))

        logger.info(
            "{} took {} to run".format(
                optimizer_instance.paths.phase_name, execution_time
            )
        )
        with open(
            "{}/execution_time".format(optimizer_instance.paths.phase_output_path), "w+"
        ) as f:
            f.write(execution_time)
        return result

    return timed_function



import inspect
import typing

import autofit as af

# noinspection PyAbstractClass
from autofit import Paths
from autofit.optimize.non_linear.multi_nest_output import MultiNestOutput


class MockNonLinearOptimizer(MultiNestOutput):
    def __init__(
        self,
        phase_name,
        phase_tag=None,
        phase_folders=tuple(),
        most_probable=None,
        model_mapper=None,
        most_likely=None,
        model_upper_params=None,
        model_lower_params=None,
    ):
        super(MockNonLinearOptimizer, self).__init__(
            model_mapper or af.ModelMapper(),
            Paths(
                phase_name=phase_name, phase_tag=phase_tag, phase_folders=phase_folders
            ),
        )

        self.most_probable = most_probable
        self.most_likely = most_likely
        self.model_upper_params = model_upper_params
        self.model_lower_params = model_lower_params

    @property
    def most_probable_model_parameters(self):
        """
        Read the most probable or most likely model values from the 'obj_summary.txt'
        file which nlo from a multinest lensing.

        This file stores the parameters of the most probable model in the first half
        of entries and the most likely model in the second half of entries. The
        offset parameter is used to start at the desiredaf.

        """
        return self.most_probable

    @property
    def most_likely_model_parameters(self):
        """
        Read the most probable or most likely model values from the 'obj_summary.txt'
        file which nlo from a \ multinest lensing.

        This file stores the parameters of the most probable model in the first half
        of entries and the most likely model in the second half of entries. The
        offset parameter is used to start at the desiredaf.
        """
        return self.most_likely

    def model_parameters_at_upper_sigma_limit(self, sigma_limit):
        return self.model_upper_params

    def model_parameters_at_lower_sigma_limit(self, sigma_limit):
        return self.model_lower_params


class MockClassNLOx4(object):
    def __init__(self, one=1, two=2, three=3, four=4):
        self.one = one
        self.two = two
        self.three = three
        self.four = four


class MockClassNLOx5(object):
    def __init__(self, one=1, two=2, three=3, four=4, five=5):
        self.one = one
        self.two = two
        self.three = three
        self.four = four
        self.five = five


class MockClassNLOx6(object):
    def __init__(self, one=(1, 2), two=(3, 4), three=3, four=4):
        self.one = one
        self.two = two
        self.three = three
        self.four = four


class MockAnalysis(object):
    def __init__(self):
        self.kwargs = None
        self.instance = None
        self.visualize_instance = None

    def fit(self, instance):
        self.instance = instance
        return 1.0

    # noinspection PyUnusedLocal
    def visualize(self, instance, *args, **kwargs):
        self.visualize_instance = instance


class SimpleClass(object):
    def __init__(self, one, two: float):
        self.one = one
        self.two = two


class ComplexClass(object):
    def __init__(self, simple: SimpleClass):
        self.simple = simple


class ListClass(object):
    def __init__(self, ls: list):
        self.ls = ls


class Distance(af.DimensionType):
    pass


class DistanceClass:
    @af.map_types
    def __init__(self, first: Distance, second: Distance):
        self.first = first
        self.second = second


class PositionClass:
    @af.map_types
    def __init__(self, position: typing.Tuple[Distance, Distance]):
        self.position = position


class DeferredClass:
    def __init__(self, one, two):
        self.one = one
        self.two = two


class Galaxy(object):
    def __init__(
        self,
        light_profiles: list = None,
        mass_profiles: list = None,
        redshift=None,
        **kwargs
    ):
        self.redshift = redshift
        self.light_profiles = light_profiles
        self.mass_profiles = mass_profiles
        self.kwargs = kwargs


class RelativeWidth(object):
    def __init__(self, one, two, three):
        self.one = one
        self.two = two
        self.three = three


class Redshift(object):
    def __init__(self, redshift):
        self.redshift = redshift


# noinspection PyAbstractClass
class GalaxyModel(af.AbstractPriorModel):
    def instance_for_arguments(self, arguments):
        try:
            return Galaxy(redshift=self.redshift.instance_for_arguments(arguments))
        except AttributeError:
            return Galaxy()

    def __init__(self, model_redshift=False, **kwargs):
        super().__init__()
        self.redshift = af.PriorModel(Redshift) if model_redshift else None
        print(self.redshift)
        self.__dict__.update(
            {
                key: af.PriorModel(value) if inspect.isclass(value) else value
                for key, value in kwargs.items()
            }
        )

    @property
    def instance_tuples(self):
        return []

    @property
    @af.cast_collection(af.PriorNameValue)
    def unique_prior_tuples(self):
        return (
            [item for item in self.__dict__.items() if isinstance(item[1], af.Prior)]
            + [("redshift", self.redshift.redshift)]
            if self.redshift is not None
            else []
        )

    @property
    @af.cast_collection(af.PriorModelNameValue)
    def flat_prior_model_tuples(self):
        return [
            item
            for item in self.__dict__.items()
            if isinstance(item[1], af.AbstractPriorModel)
        ]


class GeometryProfile(object):
    def __init__(self, centre=(0.0, 0.0)):
        """Abstract GeometryProfile, describing an object with y, x cartesian
        coordinates """
        self.centre = centre

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class SphericalProfile(GeometryProfile):
    def __init__(self, centre=(0.0, 0.0)):
        """ Generic circular profiles class to contain functions shared by light and
        mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) coordinates of the origin of the profile.
        """
        super(SphericalProfile, self).__init__(centre)


class EllipticalProfile(SphericalProfile):
    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0):
        """ Generic elliptical profiles class to contain functions shared by light
        and mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) coordinates of the origin of the profiles
        axis_ratio : float
            Ratio of profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        """
        super(EllipticalProfile, self).__init__(centre)
        self.axis_ratio = axis_ratio
        self.phi = phi


class EllipticalLP(EllipticalProfile):
    """Generic class for an elliptical light profiles"""

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0):
        """  Abstract class for an elliptical light-profile.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) coordinates of the origin of the profiles
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        """
        super(EllipticalLP, self).__init__(centre, axis_ratio, phi)


class AbstractEllipticalSersic(EllipticalProfile):
    def __init__(
        self,
        centre=(0.0, 0.0),
        axis_ratio=1.0,
        phi=0.0,
        intensity=0.1,
        effective_radius=0.6,
        sersic_index=4.0,
    ):
        """ Abstract base class for an elliptical Sersic profile, used for computing
        its effective radius and Sersic instance.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) coordinates of the origin of the profiles
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        intensity : float
            Overall intensity normalisation in the light profiles (electrons per second)
        effective_radius : float
            The circular radius containing half the light of this model_mapper
        sersic_index : Int
            The Sersic index, which controls the light profile concentration
        """
        super(AbstractEllipticalSersic, self).__init__(centre, axis_ratio, phi)
        self.intensity = intensity
        self.effective_radius = effective_radius
        self.sersic_index = sersic_index


class MassProfile(object):
    def surface_density_func(self, eta):
        raise NotImplementedError("surface_density_at_radius should be overridden")

    def surface_density_from_grid(self, grid):
        pass
        # raise NotImplementedError("surface_density_from_grid should be overridden")

    def potential_from_grid(self, grid):
        pass
        # raise NotImplementedError("potential_from_grid should be overridden")

    def deflections_from_grid(self, grid):
        raise NotImplementedError("deflections_from_grid should be overridden")


# noinspection PyAbstractClass
class EllipticalMassProfile(EllipticalProfile, MassProfile):
    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0):
        """
        Abstract class for elliptical mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The origin of the profile
        axis_ratio : float
            Ellipse's minor-to-major axis ratio (b/a)
        phi : float
            Rotation angle of profile's ellipse counter-clockwise from positive x-axis
        """
        super(EllipticalMassProfile, self).__init__(centre, axis_ratio, phi)
        self.axis_ratio = axis_ratio
        self.phi = phi


# noinspection PyAbstractClass
class EllipticalCoredPowerLaw(EllipticalMassProfile, MassProfile):
    def __init__(
        self,
        centre=(0.0, 0.0),
        axis_ratio=1.0,
        phi=0.0,
        einstein_radius=1.0,
        slope=2.0,
        core_radius=0.01,
    ):
        """
        Represents a cored elliptical power-law density distribution

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the origin of the profiles
        axis_ratio : float
            Elliptical mass profile's minor-to-major axis ratio (b/a)
        phi : float
            Rotation angle of mass profile's ellipse counter-clockwise from positive
            x-axis
        einstein_radius : float
            Einstein radius of power-law mass profiles
        slope : float
            power-law density slope of mass profiles
        core_radius : float
            The radius of the inner core
        """
        super(EllipticalCoredPowerLaw, self).__init__(centre, axis_ratio, phi)
        self.einstein_radius = einstein_radius
        self.slope = slope
        self.core_radius = core_radius


# noinspection PyAbstractClass
class EllipticalCoredIsothermal(EllipticalCoredPowerLaw):
    def __init__(
        self,
        centre=(0.0, 0.0),
        axis_ratio=1.0,
        phi=0.0,
        einstein_radius=1.0,
        core_radius=0.05,
    ):
        """
        Represents a cored elliptical isothermal density distribution, which is
        equivalent to the elliptical power-law
        density distribution for the value slope=2.0

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the origin of the profiles
        axis_ratio : float
            Elliptical mass profile's minor-to-major axis ratio (b/a)
        phi : float
            Rotation angle of mass profile's ellipse counter-clockwise from positive
            x-axis
        einstein_radius : float
            Einstein radius of power-law mass profiles
        core_radius : float
            The radius of the inner core
        """

        super(EllipticalCoredIsothermal, self).__init__(
            centre, axis_ratio, phi, einstein_radius, 2.0, core_radius
        )


class EllipticalSersic(AbstractEllipticalSersic, EllipticalLP):
    def __init__(
        self,
        centre=(0.0, 0.0),
        axis_ratio=1.0,
        phi=0.0,
        intensity=0.1,
        effective_radius=0.6,
        sersic_index=4.0,
    ):
        """ The elliptical Sersic profile, used for fitting a model_galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) origin of the light profile.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi : float
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per
            second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : Int
            Controls the concentration of the of the light profile.
        """
        super(EllipticalSersic, self).__init__(
            centre, axis_ratio, phi, intensity, effective_radius, sersic_index
        )


class EllipticalCoreSersic(EllipticalSersic):
    def __init__(
        self,
        centre=(0.0, 0.0),
        axis_ratio=1.0,
        phi=0.0,
        intensity=0.1,
        effective_radius=0.6,
        sersic_index=4.0,
        radius_break=0.01,
        intensity_break=0.05,
        gamma=0.25,
        alpha=3.0,
    ):
        """ The elliptical cored-Sersic profile, used for fitting a model_galaxy's
        light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) origin of the light profile.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi : float
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per
            second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : Int
            Controls the concetration of the of the light profile.
        radius_break : Float
            The break radius separating the inner power-law (with logarithmic slope
            gamma) and outer Sersic function.
        intensity_break : Float
            The intensity at the break radius.
        gamma : Float
            The logarithmic power-law slope of the inner core profiles
        alpha :
            Controls the sharpness of the transition between the inner core / outer
            Sersic profiles.
        """
        super(EllipticalCoreSersic, self).__init__(
            centre, axis_ratio, phi, intensity, effective_radius, sersic_index
        )
        self.radius_break = radius_break
        self.intensity_break = intensity_break
        self.alpha = alpha
        self.gamma = gamma


class EllipticalExponential(EllipticalSersic):
    def __init__(
        self,
        centre=(0.0, 0.0),
        axis_ratio=1.0,
        phi=0.0,
        intensity=0.1,
        effective_radius=0.6,
    ):
        """ The elliptical exponential profile, used for fitting a model_galaxy's light.

        This is a subset of the elliptical Sersic profile, specific to the case that
        sersic_index = 1.0.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) origin of the light profile.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi : float
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per
            second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        """
        super(EllipticalExponential, self).__init__(
            centre, axis_ratio, phi, intensity, effective_radius, 1.0
        )


class EllipticalGaussian(EllipticalLP):
    def __init__(
        self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=0.1, sigma=0.01
    ):
        """ The elliptical Gaussian profile.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) origin of the light profile.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi : float
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per
            second).
        sigma : float
            The full-width half-maximum of the Gaussian.
        """
        super(EllipticalGaussian, self).__init__(centre, axis_ratio, phi)

        self.intensity = intensity
        self.sigma = sigma


class Tracer:
    def __init__(self, lens_galaxy: Galaxy, source_galaxy: Galaxy, grid):
        self.lens_galaxy = lens_galaxy
        self.source_galaxy = source_galaxy
        self.grid = grid


class Result:
    def __init__(self, instance=None, model=None):
        self.instance = instance
        self.model = model


class HyperGalaxy(object):
    pass



import os
from functools import wraps

from autofit import conf
from autofit.mapper import link


def make_path(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        full_path = func(*args, **kwargs)
        if not os.path.exists(full_path):
            try:
                os.makedirs(full_path)
            except FileExistsError:
                pass
        return full_path

    return wrapper


def convert_paths(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if len(args) > 1:
            raise AssertionError(
                "Only phase name is allowed to be a positional argument in a phase constructor"
            )

        first_arg = kwargs.pop("paths", None)
        if first_arg is None and len(args) == 1:
            first_arg = args[0]

        if isinstance(first_arg, Paths):
            return func(self, paths=first_arg, **kwargs)

        if first_arg is None:
            first_arg = kwargs.pop("phase_name", None)

        func(
            self,
            paths=Paths(
                phase_name=first_arg,
                phase_tag=kwargs.pop("phase_tag", None),
                phase_folders=kwargs.pop("phase_folders", tuple()),
                phase_path=kwargs.pop("phase_path", None),
            ),
            **kwargs
        )

    return wrapper


class Paths:
    def __init__(
        self, phase_name="", phase_tag=None, phase_folders=tuple(), phase_path=None
    ):
        if not isinstance(phase_name, str):
            raise ValueError("Phase name must be a string")
        self.phase_path = phase_path or "/".join(phase_folders)
        self.phase_name = phase_name
        self.phase_tag = phase_tag or ""

    @property
    def path(self):
        return link.make_linked_folder(self.sym_path)

    def __eq__(self, other):
        return isinstance(other, Paths) and all(
            [
                self.phase_path == other.phase_path,
                self.phase_name == other.phase_name,
                self.phase_tag == other.phase_tag,
            ]
        )

    @property
    def phase_folders(self):
        return self.phase_path.split("/")

    @property
    def backup_path(self) -> str:
        """
        The path to the backed up optimizer folder.
        """
        return "/".join(
            filter(
                lambda item: len(item) > 0,
                [
                    conf.instance.output_path,
                    self.phase_path,
                    self.phase_name,
                    self.phase_tag,
                    "optimizer_backup",
                ],
            )
        )

    @property
    @make_path
    def phase_output_path(self) -> str:
        """
        The path to the output information for a phase.
        """
        return "{}/{}/{}/{}/".format(
            conf.instance.output_path, self.phase_path, self.phase_name, self.phase_tag
        )

    @property
    def sym_path(self) -> str:
        return "{}/{}/{}/{}/optimizer".format(
            conf.instance.output_path, self.phase_path, self.phase_name, self.phase_tag
        )

    @property
    def file_param_names(self) -> str:
        return "{}/{}".format(self.path, "multinest.paramnames")

    @property
    def file_model_info(self) -> str:
        return "{}/{}".format(self.phase_output_path, "model.info")

    @property
    @make_path
    def image_path(self) -> str:
        """
        The path to the directory in which images are stored.
        """
        return "{}image/".format(self.phase_output_path)

    @property
    @make_path
    def pdf_path(self) -> str:
        """
        The path to the directory in which images are stored.
        """
        return "{}pdf/".format(self.image_path)

    def make_optimizer_pickle_path(self) -> str:
        """
        Create the path at which the optimizer pickle should be saved
        """
        return "{}/optimizer.pickle".format(self.make_path())

    def make_model_pickle_path(self):
        """
        Create the path at which the model pickle should be saved
        """
        return "{}/model.pickle".format(self.make_path())

    @make_path
    def make_path(self) -> str:
        """
        Create the path to the folder at which the metadata and optimizer pickle should
        be saved
        """
        return "{}/{}/{}/{}/".format(
            conf.instance.output_path, self.phase_path, self.phase_name, self.phase_tag
        )

    @property
    def file_summary(self) -> str:
        return "{}/{}".format(self.backup_path, "multinestsummary.txt")

    @property
    def file_weighted_samples(self):
        return "{}/{}".format(self.backup_path, "multinest.txt")

    @property
    def file_results(self):
        return "{}/{}".format(self.phase_output_path, "model.results")



import os
import shutil
import autofit as af

path = "{}/".format(os.path.dirname(os.path.realpath(__file__)))

test_data_path = "{}/../test_files/arrays/".format(
    os.path.dirname(os.path.realpath(__file__))
)


class TestMakeAndReturnPath:
    def test__1_directory_input__makes_directory__returns_path(self):
        path = af.path_util.make_and_return_path_from_path_and_folder_names(
            path=test_data_path, folder_names=["test1"]
        )

        assert path == test_data_path + "test1/"
        assert os.path.exists(path=test_data_path + "test1")

        shutil.rmtree(test_data_path + "test1")

    def test__multiple_directories_input__makes_directory_structure__returns_full_path(
        self
    ):
        path = af.path_util.make_and_return_path_from_path_and_folder_names(
            path=test_data_path, folder_names=["test1", "test2", "test3"]
        )

        assert path == test_data_path + "test1/test2/test3/"
        assert os.path.exists(path=test_data_path + "test1")
        assert os.path.exists(path=test_data_path + "test1/test2")
        assert os.path.exists(path=test_data_path + "test1/test2/test3")

        shutil.rmtree(test_data_path + "test1")



from autofit.mapper.prior_model.recursion import DynamicRecursionCache


class Wrapper:
    def __init__(self, item):
        self.item = item


class A:
    def __init__(self, b=None):
        self.b = b


class B:
    def __init__(self, a=None):
        self.a = a


@DynamicRecursionCache()
def dict_recurse(item):
    try:
        for key, value in item.__dict__.items():
            setattr(item, key, dict_recurse(value))
    except AttributeError:
        pass
    return Wrapper(item)


def test_basic():
    a = A(B())
    a.b.a = a
    result = dict_recurse(a)
    assert isinstance(result.item.b.item.a, Wrapper)



import logging

import numpy as np
import pymultinest

from autofit import conf, exc
from autofit.optimize.non_linear.multi_nest_output import MultiNestOutput
from autofit.optimize.non_linear.non_linear import NonLinearOptimizer
from autofit.optimize.non_linear.non_linear import Result
from autofit.optimize.non_linear.non_linear import persistent_timer

logger = logging.getLogger(__name__)


class MultiNest(NonLinearOptimizer):
    def __init__(self, paths, sigma_limit=3, run=pymultinest.run):
        """
        Class to setup and run a MultiNest lensing and output the MultiNest nlo.

        This interfaces with an input model_mapper, which is used for setting up the \
        individual model instances that are passed to each iteration of MultiNest.
        """

        super().__init__(paths)

        self.sigma_limit = sigma_limit

        self.importance_nested_sampling = self.config(
            "importance_nested_sampling", bool
        )
        self.multimodal = self.config("multimodal", bool)
        self.const_efficiency_mode = self.config("const_efficiency_mode", bool)
        self.n_live_points = self.config("n_live_points", int)
        self.evidence_tolerance = self.config("evidence_tolerance", float)
        self.sampling_efficiency = self.config("sampling_efficiency", float)
        self.n_iter_before_update = self.config("n_iter_before_update", int)
        self.null_log_evidence = self.config("null_log_evidence", float)
        self.max_modes = self.config("max_modes", int)
        self.mode_tolerance = self.config("mode_tolerance", float)
        self.outputfiles_basename = self.config("outputfiles_basename", str)
        self.seed = self.config("seed", int)
        self.verbose = self.config("verbose", bool)
        self.resume = self.config("resume", bool)
        self.context = self.config("context", int)
        self.write_output = self.config("write_output", bool)
        self.log_zero = self.config("log_zero", float)
        self.max_iter = self.config("max_iter", int)
        self.init_MPI = self.config("init_MPI", bool)
        self.run = run

        logger.debug("Creating MultiNest NLO")

    def copy_with_name_extension(self, extension, remove_phase_tag=False):
        copy = super().copy_with_name_extension(
            extension=extension, remove_phase_tag=remove_phase_tag
        )
        copy.sigma_limit = self.sigma_limit
        copy.run = self.run
        copy.importance_nested_sampling = self.importance_nested_sampling
        copy.multimodal = self.multimodal
        copy.const_efficiency_mode = self.const_efficiency_mode
        copy.n_live_points = self.n_live_points
        copy.evidence_tolerance = self.evidence_tolerance
        copy.sampling_efficiency = self.sampling_efficiency
        copy.n_iter_before_update = self.n_iter_before_update
        copy.null_log_evidence = self.null_log_evidence
        copy.max_modes = self.max_modes
        copy.mode_tolerance = self.mode_tolerance
        copy.outputfiles_basename = self.outputfiles_basename
        copy.seed = self.seed
        copy.verbose = self.verbose
        copy.resume = self.resume
        copy.context = self.context
        copy.write_output = self.write_output
        copy.log_zero = self.log_zero
        copy.max_iter = self.max_iter
        copy.init_MPI = self.init_MPI
        return copy

    class Fitness(NonLinearOptimizer.Fitness):
        def __init__(
            self, nlo, analysis, instance_from_physical_vector, output_results
        ):
            super().__init__(nlo, analysis)
            self.instance_from_physical_vector = instance_from_physical_vector
            self.output_results = output_results
            self.accepted_samples = 0

            self.number_of_accepted_samples_between_output = conf.instance.general.get(
                "output", "number_of_accepted_samples_between_output", int
            )
            self.stagger_resampling_likelihood = conf.instance.non_linear.get(
                "MultiNest", "stagger_resampling_likelihood", bool
            )
            self.stagger_resampling_value = conf.instance.non_linear.get(
                "MultiNest", "stagger_resampling_value", float
            )
            self.resampling_likelihood = conf.instance.non_linear.get(
                "MultiNest", "null_log_evidence", float
            )
            self.stagger_accepted_samples = 0

        def __call__(self, cube, ndim, nparams, lnew):

            try:
                instance = self.instance_from_physical_vector(cube)
                likelihood = self.fit_instance(instance)
            except exc.FitException:

                if not self.stagger_resampling_likelihood:
                    likelihood = -np.inf
                else:

                    if self.stagger_accepted_samples < 10:

                        self.stagger_accepted_samples += 1
                        self.resampling_likelihood += self.stagger_resampling_value
                        likelihood = self.resampling_likelihood

                    else:

                        likelihood = -1.0 * np.abs(self.resampling_likelihood) * 10.0

            if likelihood > self.max_likelihood:

                self.accepted_samples += 1

                if (
                    self.accepted_samples
                    == self.number_of_accepted_samples_between_output
                ):
                    self.accepted_samples = 0
                    self.output_results(during_analysis=True)

            return likelihood

    @persistent_timer
    def fit(self, analysis, model):
        multinest_output = MultiNestOutput(model, self.paths)

        multinest_output.save_model_info()

        # noinspection PyUnusedLocal
        def prior(cube, ndim, nparams):

            # NEVER EVER REFACTOR THIS LINE!

            phys_cube = model.physical_vector_from_hypercube_vector(hypercube_vector=cube)

            for i in range(len(phys_cube)):
                cube[i] = phys_cube[i]

            return cube

        fitness_function = MultiNest.Fitness(
            self,
            analysis,
            model.instance_from_physical_vector,
            multinest_output.output_results,
        )

        logger.info("Running MultiNest...")
        self.run(
            fitness_function.__call__,
            prior,
            model.prior_count,
            outputfiles_basename="{}/multinest".format(self.paths.path),
            n_live_points=self.n_live_points,
            const_efficiency_mode=self.const_efficiency_mode,
            importance_nested_sampling=self.importance_nested_sampling,
            evidence_tolerance=self.evidence_tolerance,
            sampling_efficiency=self.sampling_efficiency,
            null_log_evidence=self.null_log_evidence,
            n_iter_before_update=self.n_iter_before_update,
            multimodal=self.multimodal,
            max_modes=self.max_modes,
            mode_tolerance=self.mode_tolerance,
            seed=self.seed,
            verbose=self.verbose,
            resume=self.resume,
            context=self.context,
            write_output=self.write_output,
            log_zero=self.log_zero,
            max_iter=self.max_iter,
            init_MPI=self.init_MPI,
        )
        logger.info("MultiNest complete")

        self.backup()
        instance = multinest_output.most_likely_model_instance
        analysis.visualize(instance=instance, during_analysis=False)
        multinest_output.output_results(during_analysis=False)
        multinest_output.output_pdf_plots()
        return Result(
            instance=instance,
            figure_of_merit=multinest_output.maximum_likelihood,
            previous_model=model,
            gaussian_tuples=multinest_output.gaussian_priors_at_sigma_limit(
                self.sigma_limit
            ),
        )



import os

import pytest

import autofit as af
import autofit.mapper.prior_model.abstract
import autofit.mapper.prior_model.prior_model
import autofit.optimize.non_linear.non_linear
from autofit import Paths
from test_autofit.mock import Galaxy, GalaxyModel

directory = os.path.dirname(os.path.realpath(__file__))

af.conf.instance = af.conf.Config(
    "{}/../../workspace/config".format(directory),
    "{}/../../workspace/output/".format(directory),
)


class NLO(autofit.optimize.non_linear.non_linear.NonLinearOptimizer):
    def fit(self, analysis, model):
        class Fitness(object):
            def __init__(self, instance_from_physical_vector, instance):
                self.result = None
                self.instance_from_physical_vector = instance_from_physical_vector
                self.instance = instance

            def __call__(self, vector):
                instance = self.instance_from_physical_vector(vector)
                for key, value in self.instance.__dict__.items():
                    setattr(instance, key, value)

                likelihood = analysis.fit(instance)
                self.result = autofit.optimize.non_linear.non_linear.Result(
                    instance, likelihood
                )

                # Return Chi squared
                return -2 * likelihood

        fitness_function = Fitness(
            model.instance_from_physical_vector, model.instance_from_prior_medians
        )
        fitness_function(model.prior_count * [0.5])

        return fitness_function.result


@pytest.fixture(name="phase")
def make_phase():
    return MyPhase(Paths(phase_name=""), optimizer_class=NLO)


class MyPhase(af.AbstractPhase):
    prop = af.PhaseProperty("prop")


@pytest.fixture(name="list_phase")
def make_list_phase():
    return MyPhase(Paths(phase_name=""), optimizer_class=NLO)


class TestPhasePropertyList(object):
    def test_classes(self, list_phase):
        objects = [GalaxyModel(), GalaxyModel()]

        list_phase.prop = objects

        assert list_phase.model.prop == objects
        assert list_phase.prop == objects

    def test_abstract_prior_models(self, list_phase):
        objects = [af.AbstractPriorModel(), af.AbstractPriorModel()]

        list_phase.prop = objects

        assert list_phase.model.prop == objects
        assert list_phase.prop == objects

    def test_mix(self, list_phase):
        objects = [GalaxyModel(), Galaxy()]

        list_phase.prop = objects

        assert list_phase.model.prop == objects
        assert list_phase.prop == objects

    def test_set_item(self, list_phase):
        galaxy_prior_0 = GalaxyModel()
        objects = [galaxy_prior_0, Galaxy()]

        list_phase.prop = objects

        galaxy_prior_1 = GalaxyModel()
        list_phase.prop[1] = galaxy_prior_1

        assert list_phase.model.prop == [galaxy_prior_0, galaxy_prior_1]

        galaxy = Galaxy()

        list_phase.prop[0] = galaxy
        assert list_phase.prop == [galaxy, galaxy_prior_1]


class TestPhasePropertyCollectionAttributes(object):
    def test_set_list_as_dict(self, list_phase):
        galaxy_model = GalaxyModel()
        list_phase.prop = dict(one=galaxy_model)

        assert len(list_phase.prop) == 1
        # noinspection PyUnresolvedReferences
        assert list_phase.prop.one == galaxy_model

    def test_override_property(self, list_phase):
        galaxy_model = GalaxyModel()

        list_phase.prop = dict(one=GalaxyModel())

        list_phase.prop.one = galaxy_model

        assert len(list_phase.prop) == 1
        assert list_phase.prop.one == galaxy_model

    def test_named_list_items(self, list_phase):
        galaxy_model = GalaxyModel()
        list_phase.prop = [galaxy_model]

        # noinspection PyUnresolvedReferences
        assert getattr(list_phase.prop, "0") == galaxy_model

    def test_mix(self, list_phase):
        objects = dict(one=GalaxyModel(), two=Galaxy())

        list_phase.prop = objects

        list_phase.prop.one = Galaxy()

        assert len(list_phase.model.prop) == 2

    def test_named_attributes_in_model(self, list_phase):
        galaxy_model = GalaxyModel(model_redshift=True)
        list_phase.prop = dict(one=galaxy_model)

        assert list_phase.model.prior_count == 1
        assert list_phase.model.prop.one == galaxy_model

        instance = list_phase.model.instance_from_prior_medians()

        assert instance.prop.one is not None
        assert len(instance.prop) == 1

    def test_named_attributes_in_model_override(self, list_phase):
        list_phase.prop = dict(one=GalaxyModel())

        assert list_phase.model.prior_count == 0

        galaxy_model = GalaxyModel(model_redshift=True)

        list_phase.prop.one = galaxy_model

        assert list_phase.model.prior_count == 1
        # assert list_phase.model.one == galaxy_model

        instance = list_phase.model.instance_from_prior_medians()

        # assert instance.one is not None
        assert len(instance.prop) == 1

    def test_named_attributes_in_instance(self, list_phase):
        galaxy = Galaxy()
        list_phase.prop = dict(one=galaxy)

        assert list_phase.model.prior_count == 0
        assert list_phase.model.prop.one == galaxy

    # def test_singular_model_info(self, list_phase):
    #     galaxy_model = test_autoarray.mock.GalaxyModel(model_redshift=True)
    #     list_phase.prop = dict(one=galaxy_model)
    #
    #     assert list_phase.model.prop.one == galaxy_model
    #     assert len(galaxy_model.flat_prior_model_tuples) == 1
    #     assert len(galaxy_model.prior_tuples) == 1
    #
    #     assert len(list_phase.model.flat_prior_model_tuples) == 1
    #
    #     print(list_phase.model.info)
    #
    #     assert len(list_phase.model.info.split('\n')) == 7

    def test_shared_priors(self, list_phase):
        list_phase.prop = dict(
            one=GalaxyModel(model_redshift=True), two=GalaxyModel(model_redshift=True)
        )

        assert list_phase.model.prior_count == 2

        # noinspection PyUnresolvedReferences
        list_phase.prop.one.redshift = list_phase.prop.two.redshift

        assert list_phase.model.prior_count == 1

    def test_hasattr(self, list_phase):
        list_phase.prop = dict()

        assert not hasattr(list_phase.prop, "one")
        list_phase.prop = dict(one=GalaxyModel(model_redshift=True))

        assert hasattr(list_phase.prop, "one")

    def test_position_not_a_prior(self, list_phase):
        list_phase.prop = [af.PriorModel(Galaxy)]

        assert list_phase.model.prior_count == 1
        assert "redshift" == list_phase.model.prior_tuples_ordered_by_id[0][0]

        prior_model = af.PriorModel(Galaxy)
        prior_model.phase_property_position = 0

        print(prior_model.instance_tuples)
        assert len(prior_model.instance_tuples) == 0



import numpy as np
import scipy.optimize

from autofit import exc
from autofit.optimize.non_linear.multi_nest_output import Output
from autofit.optimize.non_linear.non_linear import NonLinearOptimizer, persistent_timer
from autofit.optimize.non_linear.non_linear import logger


class DownhillSimplex(NonLinearOptimizer):
    def __init__(self, paths, fmin=scipy.optimize.fmin):

        super().__init__(paths)

        self.xtol = self.config("xtol", float)
        self.ftol = self.config("ftol", float)
        self.maxiter = self.config("maxiter", int)
        self.maxfun = self.config("maxfun", int)

        self.full_output = self.config("full_output", int)
        self.disp = self.config("disp", int)
        self.retall = self.config("retall", int)

        self.fmin = fmin

        logger.debug("Creating DownhillSimplex NLO")

    def copy_with_name_extension(self, extension, remove_phase_tag=False):
        copy = super().copy_with_name_extension(
            extension=extension, remove_phase_tag=remove_phase_tag
        )
        copy.fmin = self.fmin
        copy.xtol = self.xtol
        copy.ftol = self.ftol
        copy.maxiter = self.maxiter
        copy.maxfun = self.maxfun
        copy.full_output = self.full_output
        copy.disp = self.disp
        copy.retall = self.retall
        return copy

    class Fitness(NonLinearOptimizer.Fitness):
        def __init__(self, nlo, analysis, instance_from_physical_vector):
            super().__init__(nlo, analysis)
            self.instance_from_physical_vector = instance_from_physical_vector

        def __call__(self, vector):
            try:
                instance = self.instance_from_physical_vector(vector)
                likelihood = self.fit_instance(instance)
            except exc.FitException:
                likelihood = -np.inf
            return -2 * likelihood

    @persistent_timer
    def fit(self, analysis, model):
        dhs_output = Output(model, self.paths)
        dhs_output.save_model_info()
        initial_vector = model.physical_values_from_prior_medians

        fitness_function = DownhillSimplex.Fitness(
            self, analysis, model.instance_from_physical_vector
        )

        logger.info("Running DownhillSimplex...")
        output = self.fmin(fitness_function, x0=initial_vector)
        logger.info("DownhillSimplex complete")

        self.backup()
        res = fitness_function.result

        # Create a set of Gaussian priors from this result and associate them with the result object.
        res.gaussian_tuples = [(mean, 0) for mean in output]
        res.previous_model = model

        analysis.visualize(instance=res.instance, during_analysis=False)
        return res



import math

import autofit as af
from autofit.optimize.non_linear.multi_nest_output import Output


class MockNLO(af.NonLinearOptimizer):
    def fit(self, analysis, model):
        mock_output = Output(model, self.paths)
        mock_output.save_model_info()
        if model.prior_count == 0:
            raise AssertionError("There are no priors associated with the model!")
        if model.prior_count != len(model.unique_prior_paths):
            raise AssertionError(
                "Prior count doesn't match number of unique prior paths"
            )
        index = 0
        unit_vector = model.prior_count * [0.5]
        while True:
            try:
                instance = model.instance_from_unit_vector(unit_vector)
                fit = analysis.fit(instance)
                break
            except af.exc.FitException as e:
                unit_vector[index] += 0.1
                if unit_vector[index] >= 1:
                    raise e
                index = (index + 1) % model.prior_count
        return af.Result(
            instance,
            fit,
            model,
            gaussian_tuples=[
                (prior.mean, prior.width if math.isfinite(prior.width) else 1.0)
                for prior in sorted(model.priors, key=lambda prior: prior.id)
            ],
        )


class MockAnalysis(af.Analysis):
    def fit(self, instance):
        return 1.0

    def visualize(self, instance, during_analysis):
        pass

    def __init__(self, data, results):
        self.data = data
        self.results = results



import copy

import pytest

import autofit as af
from autofit import Paths
from test_autofit import mock


@pytest.fixture(name="phase")
def make_phase():
    phase = af.AbstractPhase(phase_name="phase name")
    phase.model.one = af.PriorModel(mock.Galaxy, light=mock.EllipticalLP)
    return phase


@pytest.fixture(name="model_promise")
def make_model_promise(phase):
    return phase.result.model.one.redshift


@pytest.fixture(name="instance_promise")
def make_instance_promise(phase):
    return phase.result.instance.one.redshift


@pytest.fixture(name="profile_promise")
def make_profile_promise(phase):
    return phase.result.model.one.light


@pytest.fixture(name="collection")
def make_collection():
    collection = af.ResultsCollection()
    model = af.ModelMapper()
    model.one = af.PriorModel(mock.Galaxy, light=mock.EllipticalLP)
    instance = af.ModelInstance()
    instance.one = mock.Galaxy(light=mock.EllipticalLP())

    result = mock.Result(model=model, instance=instance)

    model = af.ModelMapper()
    instance = af.ModelInstance()

    model.hyper_galaxy = mock.HyperGalaxy
    instance.hyper_galaxy = mock.HyperGalaxy()

    hyper_result = mock.Result(model=model, instance=instance)

    result.hyper_result = hyper_result

    collection.add("phase name", result)

    return collection


@pytest.fixture(name="last_model")
def make_last_model():
    return af.last.model.one.redshift


@pytest.fixture(name="last_instance")
def make_last_instance():
    return af.last.instance.one.redshift


class TestLastPromises:
    def test_model(self, last_model):
        assert last_model.path == ("one", "redshift")
        assert last_model.is_instance is False

    def test_instance(self, last_instance):
        assert last_instance.path == ("one", "redshift")
        assert last_instance.is_instance is True

    def test_recover_model(self, collection, last_model):
        result = last_model.populate(collection)

        assert result is collection[0].model.one.redshift

    def test_recover_instance(self, collection, last_instance):
        result = last_instance.populate(collection)

        assert result is collection[0].instance.one.redshift

    def test_recover_last_model(self, collection, last_model):
        last_results = copy.deepcopy(collection.last)
        collection.add("last_phase", last_results)

        result = last_model.populate(collection)
        assert result is last_results.model.one.redshift
        assert result is not collection[0].model.one.redshift

    def test_embedded_results(self, collection):
        hyper_result = af.last.hyper_result

        model_promise = hyper_result.model
        instance_promise = hyper_result.instance

        model = model_promise.populate(collection)
        instance = instance_promise.populate(collection)

        assert isinstance(model.hyper_galaxy, af.PriorModel)
        assert model.hyper_galaxy.cls is mock.HyperGalaxy
        assert isinstance(instance.hyper_galaxy, mock.HyperGalaxy)

    def test_raises(self, collection):
        bad_promise = af.last.model.a.bad.path
        with pytest.raises(AttributeError):
            bad_promise.populate(collection)


class TestCase:
    def test_model_promise(self, model_promise, phase):
        assert isinstance(model_promise, af.Promise)
        assert model_promise.path == ("one", "redshift")
        assert model_promise.is_instance is False
        assert model_promise.phase is phase

    def test_instance_promise(self, instance_promise, phase):
        assert isinstance(instance_promise, af.Promise)
        assert instance_promise.path == ("one", "redshift")
        assert instance_promise.is_instance is True
        assert instance_promise.phase is phase

    def test_non_existent(self, phase):
        with pytest.raises(AttributeError):
            assert phase.result.model.one.bad

        with pytest.raises(AttributeError):
            assert phase.result.instance.one.bad

    def test_recover_model(self, collection, model_promise):
        result = model_promise.populate(collection)

        assert result is collection[0].model.one.redshift

    def test_recover_instance(self, collection, instance_promise):
        result = instance_promise.populate(collection)

        assert result is collection[0].instance.one.redshift

    def test_populate_prior_model_model(self, collection, model_promise):
        new_galaxy = af.PriorModel(mock.Galaxy, redshift=model_promise)

        result = new_galaxy.populate(collection)

        assert result.redshift is collection[0].model.one.redshift

    def test_populate_prior_model_instance(self, collection, instance_promise):
        new_galaxy = af.PriorModel(mock.Galaxy, redshift=instance_promise)

        result = new_galaxy.populate(collection)

        assert result.redshift is collection[0].instance.one.redshift

    def test_kwarg_promise(self, profile_promise, collection):
        galaxy = af.PriorModel(mock.Galaxy, light=profile_promise)
        populated = galaxy.populate(collection)

        assert isinstance(populated.light, af.PriorModel)

        instance = populated.instance_from_prior_medians()

        assert isinstance(instance.kwargs["light"], mock.EllipticalLP)

    def test_embedded_results(self, phase, collection):
        hyper_result = phase.result.hyper_result

        assert isinstance(hyper_result, af.PromiseResult)

        model_promise = hyper_result.model
        instance_promise = hyper_result.instance

        print(model_promise.path)

        assert isinstance(model_promise.hyper_galaxy, af.Promise)
        assert isinstance(instance_promise.hyper_galaxy, af.Promise)

        model = model_promise.populate(collection)
        instance = instance_promise.populate(collection)

        assert isinstance(model.hyper_galaxy, af.PriorModel)
        assert model.hyper_galaxy.cls is mock.HyperGalaxy
        assert isinstance(instance.hyper_galaxy, mock.HyperGalaxy)



from functools import wraps


class RecursionPromise:
    pass


def replace_promise(promise: RecursionPromise, obj, true_value):
    """
    Traverse the object replacing any identity of the promise with the true value

    Parameters
    ----------
    promise
        A placeholder for an object that had not been computed at the time some part of the object was computed
    obj
        An object computed that may contain Promises
    true_value
        The true value associated with the promise

    Returns
    -------
    obj
        The object with any identities of the Promise replaced
    """
    if isinstance(obj, list):
        return [replace_promise(promise, item, true_value) for item in obj]
    if isinstance(obj, dict):
        return {
            key: replace_promise(promise, value, true_value)
            for key, value in obj.items()
        }
    if obj is promise:
        return true_value
    try:
        for key, value in obj.__dict__.items():
            setattr(obj, key, replace_promise(promise, value, true_value))
    except (AttributeError, TypeError):
        pass
    return obj


class DynamicRecursionCache:
    def __init__(self):
        """
        A decorating class that prevents infinite loops when recursing graphs by attaching placeholders
        """
        self.cache = dict()

    def __call__(self, func):
        """
        Decorate the function to prevent recursion.

        When the function is called with a set of arguments, A, a Promise is stored for that set of arguments in the
        cache. If the function is called again with that set of arguments then the Promise is returned. When the
        function itself returns a value any identity of the Promise is replaced by the actual value returned.
        """

        @wraps(func)
        def wrapper(item, *args, **kwargs):
            print(f"Recursion wrapper received item {item}")
            item_id = id(item)

            cache_keys = ",".join(map(str, self.cache.keys()))
            print(f"This gives item_id {item_id}. Cache keys = {cache_keys}")
            if item_id in self.cache:
                print("Item in cache")
                return self.cache[item_id]
            print("item not in cache")
            recursion_promise = RecursionPromise()
            self.cache[item_id] = recursion_promise
            result = func(item, *args, **kwargs)
            result = replace_promise(recursion_promise, result, result)
            del self.cache[item_id]
            return result

        return wrapper



import logging
import shutil
from os import path

import pytest


from autofit.optimize import grid_search as gs

import test

import autofit as af

logger = logging.getLogger(__name__)

try:
    output_dir = "{}/../../autolens_workspace/output".format(
        path.dirname(path.realpath(__file__))
    )
    logger.info("Removing {}".format(output_dir))
    shutil.rmtree(output_dir)
except FileNotFoundError:
    logging.info("Not found")


class Analysis(af.Analysis):
    def fit(self, instance):
        return -(instance.profile.centre[0] ** 2 + instance.profile.centre[1] ** 2)

    def visualize(self, instance, **kwargs):
        pass

    def describe(self, instance):
        return "{}, {}".format(*instance.profile.centre)


class Phase(af.AbstractPhase):
    profile = af.PhaseProperty("profile")
    instance_profile = af.PhaseProperty("instance_profile")

    def __init__(
        self,
        phase_name,
        phase_folders,
        profile,
        instance_profile,
        optimizer_class=af.MultiNest,
    ):
        super().__init__(
            phase_name=phase_name,
            phase_tag="phase_tag",
            phase_folders=phase_folders,
            optimizer_class=optimizer_class,
        )
        self.profile = profile
        self.instance_profile = instance_profile

    def make_result(self, result, analysis):
        return result


class TestCase(object):
    def test_integration(self):

        multinest = af.MultiNest(
            phase_folders=["integration"], phase_name="test_autoarray"
        )

        multinest.model.profile = test_autofit.mock.EllipticalProfile

        result = multinest.fit(Analysis())

        centre = result.instance.profile.centre

        assert 0 == pytest.approx(centre[0], abs=0.1)
        assert 0 == pytest.approx(centre[1], abs=0.1)

    def test_grid(self):

        grid_search = gs.GridSearch(
            phase_name="phase_grid_search",
            phase_tag="_tag",
            phase_folders=["integration"],
        )
        grid_search.model.profile = test_autofit.mock.EllipticalProfile

        # noinspection PyUnresolvedReferences
        result = grid_search.fit(
            Analysis(),
            [grid_search.model.profile.centre_0, grid_search.model.profile.centre_1],
        )

    def test_phase(self):

        phase = Phase(
            phase_name="test_phase",
            phase_folders=["integration"],
            profile=test_autofit.mock.EllipticalProfile,
            instance_profile=test_autofit.mock.EllipticalProfile(),
        )
        result = phase.run_analysis(Analysis())

        centre = result.instance.profile.centre

        assert 0 == pytest.approx(centre[0], abs=0.1)
        assert 0 == pytest.approx(centre[1], abs=0.1)

    def test_classic_grid_search_phase(self):
        # noinspection PyTypeChecker
        phase = Phase(
            phase_name="phase_classic_grid_search_phase",
            phase_folders=["integration"],
            profile=test_autofit.mock.EllipticalProfile,
            instance_profile=test_autofit.mock.EllipticalProfile(),
            optimizer_class=af.GridSearch,
        )
        result = phase.run_analysis(Analysis())

        centre = result.instance.profile.centre

        assert 0 == pytest.approx(centre[0], abs=0.1)
        assert 0 == pytest.approx(centre[1], abs=0.1)

    def test_grid_search_phase(self):
        class GridSearchPhase(af.as_grid_search(Phase)):
            @property
            def grid_priors(self):
                return [self.model.profile.centre_0, self.model.profile.centre_1]

        instance_profile = test_autofit.mock.EllipticalProfile()

        result = GridSearchPhase(
            phase_name="grid_search_phase",
            phase_folders=["integration"],
            number_of_steps=2,
            profile=test_autofit.mock.EllipticalProfile,
            instance_profile=instance_profile,
        ).run_analysis(Analysis())

        assert result.results[0].instance.instance_profile == instance_profile

        assert result.figure_of_merit_array[0, 0] > result.figure_of_merit_array[0, 1]
        assert result.figure_of_merit_array[0, 0] > result.figure_of_merit_array[1, 0]
        assert result.figure_of_merit_array[1, 0] > result.figure_of_merit_array[1, 1]
        assert result.figure_of_merit_array[0, 1] > result.figure_of_merit_array[1, 1]

    def test__grid_search_phase_parallel(self):
        class GridSearchPhase(af.as_grid_search(Phase, parallel=True)):
            @property
            def grid_priors(self):
                return [self.model.profile.centre_0, self.model.profile.centre_1]

        instance_profile = test_autofit.mock.EllipticalProfile()

        result = GridSearchPhase(
            phase_name="grid_search_phase_parallel",
            phase_folders=["integration"],
            number_of_steps=2,
            profile=test_autofit.mock.EllipticalProfile,
            instance_profile=instance_profile,
        ).run_analysis(Analysis())

        assert result.results[0].instance.instance_profile == instance_profile

        print(result.figure_of_merit_array)

        assert result.figure_of_merit_array[0, 0] > result.figure_of_merit_array[0, 1]
        assert result.figure_of_merit_array[0, 0] > result.figure_of_merit_array[1, 0]
        assert result.figure_of_merit_array[1, 0] > result.figure_of_merit_array[1, 1]
        assert result.figure_of_merit_array[0, 1] > result.figure_of_merit_array[1, 1]


if __name__ == "__main__":
    TestCase().test_integration()
    TestCase().test_grid()
    TestCase().test_phase()
    TestCase().test_classic_grid_search_phase()
    TestCase().test_grid_search_phase()



from autofit.mapper.prior_model.prior_model import PriorModel, Prior


class AnnotationPriorModel(PriorModel):
    def __init__(self, cls, parent_class, true_argument_name, **kwargs):
        self.parent_class = parent_class
        self.true_argument_name = true_argument_name
        super().__init__(cls, **kwargs)

    def make_prior(self, attribute_name):
        return Prior.for_class_and_attribute_name(
            self.parent_class, self.true_argument_name
        )



import copy
from typing import Optional, Union, Tuple

from autofit.mapper.model_object import ModelObject
from autofit.mapper.prior_model.recursion import DynamicRecursionCache
from autofit.tools.pipeline import ResultsCollection
from autofit.tools.promise import AbstractPromise


class AbstractModel(ModelObject):
    def __add__(self, other):
        instance = self.__class__()

        def add_items(item_dict):
            for key, value in item_dict.items():
                if isinstance(value, list) and hasattr(instance, key):
                    setattr(instance, key, getattr(instance, key) + value)
                else:
                    setattr(instance, key, value)

        add_items(self.__dict__)
        add_items(other.__dict__)
        return instance

    def copy(self):
        return copy.deepcopy(self)

    def populate(self, collection):
        return populate(self, collection)

    def object_for_path(self, path: (str,)) -> object:
        """
        Get the object at a given path.

        Parameters
        ----------
        path
            A tuple describing the path to an object in the model tree

        Returns
        -------
        object
            The object
        """
        instance = self
        for name in path:
            instance = getattr(instance, name)
        return instance

    def path_instance_tuples_for_class(self, cls: type, ignore_class=None):
        """
        Tuples containing the path tuple and instance for every instance of the class
        in the model tree.

        Parameters
        ----------
        ignore_class
            Children of instances of this class are ignored
        cls
            The type to find instances of

        Returns
        -------
        path_instance_tuples: [((str,), object)]
            Tuples containing the path to and instance of objects of the given type.
        """
        return path_instances_of_class(self, cls, ignore_class=ignore_class)

    def direct_tuples_with_type(self, class_type):
        return list(
            filter(
                lambda t: t[0] != "id" and isinstance(t[1], class_type),
                self.__dict__.items(),
            )
        )

    def attribute_tuples_with_type(self, class_type, ignore_class=None):
        return [
            (t[0][-1], t[1])
            for t in self.path_instance_tuples_for_class(
                class_type, ignore_class=ignore_class
            )
        ]


@DynamicRecursionCache()
def populate(obj, collection: ResultsCollection):
    """
    Replace promises with instances and instances. Promises are placeholders expressing that a given attribute should
    be replaced with an actual value once the phase that generates that value is complete.

    Parameters
    ----------
    obj
        The object to be populated
    collection
        A collection of Results from previous phases

    Returns
    -------
    obj
        The same object with all promises populated, or if the object was a promise the replacement for that promise
    """
    if isinstance(obj, list):
        return [populate(item, collection) for item in obj]
    if isinstance(obj, dict):
        return {key: populate(value, collection) for key, value in obj.items()}
    if isinstance(obj, AbstractPromise):
        return obj.populate(collection)
    try:
        new = copy.deepcopy(obj)
        for key, value in obj.__dict__.items():
            setattr(new, key, populate(value, collection))
        return new
    except (AttributeError, TypeError):
        return obj


# @DynamicRecursionCache()
def path_instances_of_class(
    obj, cls: type, ignore_class: Optional[Union[type, Tuple[type]]] = None
):
    """
    Recursively search the object for instances of a given class

    Parameters
    ----------
    obj
        The object to recursively search
    cls
        The type to search for
    ignore_class
        A type or

    Returns
    -------
    instance of type
    """
    if ignore_class is not None and isinstance(obj, ignore_class):
        return []
    if isinstance(obj, cls):
        return [(tuple(), obj)]
    results = []
    try:
        from autofit.mapper.prior_model.annotation import AnnotationPriorModel

        for key, value in obj.__dict__.items():
            for item in path_instances_of_class(value, cls, ignore_class=ignore_class):
                if isinstance(value, AnnotationPriorModel):
                    path = (key,)
                else:
                    path = (key, *item[0])
                results.append((path, item[1]))
        return results
    except (AttributeError, TypeError):
        return []


class ModelInstance(AbstractModel):
    """
    An object to hold model instances produced by providing arguments to a model mapper.

    @DynamicAttrs
    """

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __getitem__(self, item):
        return self.items[item]

    @property
    def items(self):
        return list(self.dict.values())

    @property
    def dict(self):
        return {
            key: value
            for key, value in self.__dict__.items()
            if key not in ("id", "component_number", "item_number")
        }

    def __len__(self):
        return len(self.items)

    def as_model(self, model_classes=tuple()):
        from autofit.mapper.prior_model.abstract import AbstractPriorModel

        return AbstractPriorModel.from_instance(self, model_classes)



from autofit.mapper.model import path_instances_of_class
from autofit.mapper.prior_model.prior import instanceNameValue
from autofit.mapper.prior_model.prior import (
    Prior,
    UniformPrior,
    GaussianPrior,
    LogUniformPrior,
    AttributeNameValue,
)
from autofit.mapper.prior_model.prior import PriorNameValue
from autofit.mapper.prior_model.prior import cast_collection
from . import conf
from . import exc
from .aggregator import Aggregator
from .mapper import *
from .mapper import link
from .mapper.model import AbstractModel
from .mapper.model import ModelInstance
from .mapper.model_mapper import ModelMapper
from .mapper.model_object import ModelObject
from .mapper.prior_model import *
from .mapper.prior_model.abstract import AbstractPriorModel
from .mapper.prior_model.annotation import AnnotationPriorModel
from .mapper.prior_model.collection import CollectionPriorModel
from .mapper.prior_model.deferred import DeferredArgument
from .mapper.prior_model.deferred import DeferredInstance
from .mapper.prior_model.dimension_type import DimensionType, map_types
from .mapper.prior_model.prior_model import PriorModel
from .mapper.prior_model.util import PriorModelNameValue
from .optimize import *
from .optimize.non_linear.downhill_simplex import DownhillSimplex
from .optimize.non_linear.downhill_simplex import DownhillSimplex
from .optimize.non_linear.grid_search import GridSearch
from .optimize.non_linear.multi_nest import MultiNest
from .optimize.non_linear.multi_nest import MultiNest
from .optimize.non_linear.non_linear import Analysis
from .optimize.non_linear.non_linear import NonLinearOptimizer
from .optimize.non_linear.non_linear import NonLinearOptimizer
from autofit.optimize.non_linear.paths import Paths
from autofit.optimize.non_linear.paths import convert_paths
from .optimize.non_linear.non_linear import Result
from .tools import *
from .tools import path_util, text_util
from .tools.phase import AbstractPhase
from .tools.phase import Phase
from .tools.phase import as_grid_search
from .tools.phase_property import PhaseProperty
from .tools.pipeline import Pipeline
from .tools.pipeline import ResultsCollection
from .tools.promise import Promise
from .tools.promise import PromiseResult
from .tools.promise import last

__version__ = '0.45.0'



import ast
import os

import numpy as np

from autofit import conf, exc
from autofit.optimize import optimizer as opt
from autofit.optimize.non_linear.multi_nest_output import Output
from autofit.optimize.non_linear.non_linear import (
    NonLinearOptimizer,
    Result,
    IntervalCounter,
    persistent_timer,
)
from autofit.optimize.non_linear.non_linear import logger
from autofit.optimize.non_linear.paths import Paths


class GridSearch(NonLinearOptimizer):
    def __init__(self, paths, step_size=None, grid=opt.grid):
        """
        Optimise by performing a grid search.

        Parameters
        ----------
        step_size: float | None
            The step size of the grid search in hypercube space.
            E.g. a step size of 0.5 will give steps 0.0, 0.5 and 1.0
        grid: function
            A function that takes a fitness function, dimensionality and step size and performs a grid search
        """
        super().__init__(paths)
        self.step_size = step_size or self.config("step_size", float)
        self.grid = grid

    def copy_with_name_extension(self, extension):
        name = "{}/{}".format(self.paths.phase_name, extension)

        new_instance = self.__class__(
            Paths(
                phase_name=name,
                phase_folders=self.paths.phase_folders,
                phase_tag=self.paths.phase_tag,
            ),
            step_size=self.step_size,
        )
        new_instance.grid = self.grid

        return new_instance

    class Result(Result):
        def __init__(self, result, instances, previous_model, gaussian_tuples):
            """
            The result of an grid search optimization.

            Parameters
            ----------
            result: Result
                The result
            instances: [mm.ModelInstance]
                A model instance for each point in the grid search
            """
            super().__init__(
                result.instance, result.figure_of_merit, previous_model, gaussian_tuples
            )
            self.instances = instances

        def __str__(self):
            return "Analysis Result:\n{}".format(
                "\n".join(
                    [
                        "{}: {}".format(key, value)
                        for key, value in self.__dict__.items()
                    ]
                )
            )

    class Fitness(NonLinearOptimizer.Fitness):
        def __init__(
            self,
            nlo,
            analysis,
            instance_from_unit_vector,
            save_results,
            prior_count,
            checkpoint_count=0,
            best_fit=-np.inf,
            best_cube=None,
        ):
            super().__init__(nlo, analysis)
            self.instance_from_unit_vector = instance_from_unit_vector
            self.total_calls = 0
            self.checkpoint_count = checkpoint_count
            self.save_results = save_results
            self.best_fit = best_fit
            self.best_cube = best_cube
            self.prior_count = prior_count
            self.all_fits = {}
            grid_results_interval = conf.instance.general.get(
                "output", "grid_results_interval", int
            )

            self.should_save_grid_results = IntervalCounter(grid_results_interval)
            if self.best_cube is not None:
                self.fit_instance(self.instance_from_unit_vector(self.best_cube))

        def __call__(self, cube):
            try:
                self.total_calls += 1
                if self.total_calls <= self.checkpoint_count:
                    #  TODO: is there an issue here where grid_search will forget the previous best fit?
                    return -np.inf
                instance = self.instance_from_unit_vector(cube)
                fit = self.fit_instance(instance)
                self.all_fits[cube] = fit
                if fit > self.best_fit:
                    self.best_fit = fit
                    self.best_cube = cube
                self.nlo.save_checkpoint(
                    self.total_calls, self.best_fit, self.best_cube, self.prior_count
                )
                if self.should_save_grid_results():
                    self.save_results(self.all_fits.items())
                return fit
            except exc.FitException:
                return -np.inf

    @property
    def checkpoint_path(self):
        return "{}/.checkpoint".format(self.paths.path)

    def save_checkpoint(self, total_calls, best_fit, best_cube, prior_count):
        with open(self.checkpoint_path, "w+") as f:

            def write(item):
                f.writelines("{}\n".format(item))

            write(total_calls)
            write(best_fit)
            write(best_cube)
            write(self.step_size)
            write(prior_count)

    @property
    def is_checkpoint(self):
        return os.path.exists(self.checkpoint_path)

    @property
    def checkpoint_array(self):
        with open(self.checkpoint_path) as f:
            return f.readlines()

    @property
    def checkpoint_count(self):
        return int(self.checkpoint_array[0])

    @property
    def checkpoint_fit(self):
        return float(self.checkpoint_array[1])

    @property
    def checkpoint_cube(self):
        return ast.literal_eval(self.checkpoint_array[2])

    @property
    def checkpoint_step_size(self):
        return float(self.checkpoint_array[3])

    @property
    def checkpoint_prior_count(self):
        return int(self.checkpoint_array[4])

    @persistent_timer
    def fit(self, analysis, model):
        gs_output = Output(model, self.paths)

        gs_output.save_model_info()

        checkpoint_count = 0
        best_fit = -np.inf
        best_cube = None

        if self.is_checkpoint:
            if not self.checkpoint_prior_count == model.prior_count:
                raise exc.CheckpointException(
                    "The number of dimensions does not match that found in the checkpoint"
                )
            if not self.checkpoint_step_size == self.step_size:
                raise exc.CheckpointException(
                    "The step size does not match that found in the checkpoint"
                )

            checkpoint_count = self.checkpoint_count
            best_fit = self.checkpoint_fit
            best_cube = self.checkpoint_cube

        def save_results(all_fit_items):
            results_list = [model.param_names + ["fit"]]
            for item in all_fit_items:
                results_list.append(
                    [*model.physical_vector_from_hypercube_vector(item[0]), item[1]]
                )

            with open("{}/results".format(self.paths.phase_output_path), "w+") as f:
                f.write(
                    "\n".join(
                        map(
                            lambda ls: ", ".join(
                                map(
                                    lambda value: "{:.2f}".format(value)
                                    if isinstance(value, float)
                                    else str(value),
                                    ls,
                                )
                            ),
                            results_list,
                        )
                    )
                )

        fitness_function = GridSearch.Fitness(
            self,
            analysis,
            model.instance_from_unit_vector,
            save_results,
            model.prior_count,
            checkpoint_count=checkpoint_count,
            best_fit=best_fit,
            best_cube=best_cube,
        )

        logger.info("Running grid search...")
        self.grid(fitness_function, model.prior_count, self.step_size)

        logger.info("grid search complete")
        self.backup()

        res = fitness_function.result

        instances = [
            (model.instance_from_unit_vector(cube), fit)
            for cube, fit in fitness_function.all_fits.items()
        ]

        # Create a set of Gaussian priors from this result and associate them with the result object.
        res = GridSearch.Result(
            res, instances, model, [(mean, 0) for mean in fitness_function.best_cube]
        )

        analysis.visualize(instance=res.instance, during_analysis=False)

        return res



import os
import shutil

import pytest

import autofit as af
from autofit import conf

directory = os.path.dirname(os.path.realpath(__file__))

text_path = "{}/../test_files/text/".format(os.path.dirname(os.path.realpath(__file__)))


@pytest.fixture(scope="session", autouse=True)
def do_something():
    conf.instance = conf.Config(
        config_path="{}/../test_files/config/text".format(directory)
    )


def test__label_and_label_string():
    string0 = af.text_util.label_and_label_string(
        label0="param0", label1="mass", whitespace=10
    )
    string1 = af.text_util.label_and_label_string(
        label0="param00", label1="mass0", whitespace=10
    )
    string2 = af.text_util.label_and_label_string(
        label0="param000", label1="mass111", whitespace=10
    )

    assert string0 == "param0    mass"
    assert string1 == "param00   mass0"
    assert string2 == "param000  mass111"

    string0 = af.text_util.label_and_label_string(
        label0="param0", label1="mass", whitespace=20
    )

    assert string0 == "param0              mass"


def test__label_and_value_string():
    string0 = af.text_util.label_and_value_string(
        label="param0", value=2.0, whitespace=10
    )
    string1 = af.text_util.label_and_value_string(
        label="param00", value=2.0, whitespace=10
    )
    string2 = af.text_util.label_and_value_string(
        label="param000", value=2.0, whitespace=10
    )

    assert string0 == "param0    2.00"
    assert string1 == "param00   2.00"
    assert string2 == "param000  2.00"

    string = af.text_util.label_and_value_string(
        label="param11", value=3.00, whitespace=20
    )

    assert string == "param11             3.0000"

    string = af.text_util.label_and_value_string(
        label="param12", value=3.00, whitespace=15
    )

    assert string == "param12        3.00e+00"


def test__label_value_and_limits_string():
    string0 = af.text_util.label_value_and_limits_string(
        label="param0", value=2.0, upper_limit=2.5, lower_limit=1.5, whitespace=10
    )
    string1 = af.text_util.label_value_and_limits_string(
        label="param00", value=2.0, upper_limit=2.7, lower_limit=1.3, whitespace=10
    )
    string2 = af.text_util.label_value_and_limits_string(
        label="param000", value=2.0, upper_limit=2.9, lower_limit=1.1, whitespace=10
    )

    assert string0 == "param0    2.00 (1.50, 2.50)"
    assert string1 == "param00   2.00 (1.30, 2.70)"
    assert string2 == "param000  2.00 (1.10, 2.90)"

    string = af.text_util.label_value_and_limits_string(
        label="param11",
        value=3.00,
        upper_limit=40000.0,
        lower_limit=0.0001,
        whitespace=20,
    )

    assert string == "param11             3.0000 (0.0001, 40000.0000)"

    string = af.text_util.label_value_and_limits_string(
        label="param12", value=3.00, upper_limit=500.0, lower_limit=1.0, whitespace=15
    )

    assert string == "param12        3.00e+00 (1.00e+00, 5.00e+02)"


def test__label_value_and_unit_string():
    string0 = af.text_util.label_value_and_unit_string(
        label="param0", value=2.0, unit="arcsec", whitespace=10
    )
    string1 = af.text_util.label_value_and_unit_string(
        label="param00", value=2.0, unit="mass", whitespace=10
    )
    string2 = af.text_util.label_value_and_unit_string(
        label="param000", value=2.0, unit="kg", whitespace=10
    )

    assert string0 == "param0    2.00 arcsec"
    assert string1 == "param00   2.00 mass"
    assert string2 == "param000  2.00 kg"

    string = af.text_util.label_value_and_unit_string(
        label="param11", value=3.00, unit="kg", whitespace=20
    )

    assert string == "param11             3.0000 kg"

    string = af.text_util.label_value_and_unit_string(
        label="param12", value=3.00, unit="kgs", whitespace=15
    )

    assert string == "param12        3.00e+00 kgs"


def test__output_list_of_strings_to_file():
    if os.path.exists(text_path):
        shutil.rmtree(text_path)

    os.mkdir(text_path)

    results = ["hi\n", "hello"]
    af.text_util.output_list_of_strings_to_file(
        file=text_path + "model.results", list_of_strings=results
    )

    file = open(text_path + "model.results", "r")

    assert file.readlines() == ["hi\n", "hello"]


def test__within_radius_label_value_and_unit_string():
    string0 = af.text_util.within_radius_label_value_and_unit_string(
        prefix="mass",
        radius=1.0,
        unit_length="arcsec",
        value=30.0,
        unit_value="solMass",
        whitespace=40,
    )

    string1 = af.text_util.within_radius_label_value_and_unit_string(
        prefix="mass",
        radius=1.0,
        unit_length="arcsec",
        value=30.0,
        unit_value="solMass",
        whitespace=35,
    )

    string2 = af.text_util.within_radius_label_value_and_unit_string(
        prefix="mass",
        radius=1.0,
        unit_length="arcsec",
        value=30.0,
        unit_value="solMass",
        whitespace=30,
    )

    assert string0 == "mass_within_1.00_arcsec                 mass_value 30.0"
    assert string1 == "mass_within_1.00_arcsec            mass_value 30.0"
    assert string2 == "mass_within_1.00_arcsec       mass_value 30.0"

    string = af.text_util.within_radius_label_value_and_unit_string(
        prefix="mass",
        radius=1.0,
        unit_length="arcsec2",
        value=40.0,
        unit_value="solMass2",
        whitespace=40,
    )

    assert string == "mass_within_1.00_arcsec2                mass_value 40.0"


def test_string():
    assert af.text_util.format_string_for_label("radius_value") == "radius_value"
    assert af.text_util.format_string_for_label("mass_value") == "mass_value"


def test_substring():
    assert (
        af.text_util.format_string_for_label("einstein_radius_value") == "radius_value"
    )
    assert af.text_util.format_string_for_label("mass_value_something") == "mass_value"



from os import path

import pytest

import autofit as af
from autofit import conf
import shutil

directory = path.dirname(path.realpath(__file__))


@pytest.fixture(autouse=True)
def set_config_path():
    conf.instance = conf.Config(
        path.join(directory, "test_files/config"), path.join(directory, "output")
    )


@pytest.fixture(autouse=True)
def remove_output():
    try:
        shutil.rmtree(f"{directory}/output")
    except FileNotFoundError:
        pass


@pytest.fixture
def model():
    return af.ModelMapper()



import pickle

import autofit as af
from autofit import Paths


class TestCase(object):
    def test_simple_pickle(self):
        optimiser = af.MultiNest(Paths("phasename"))
        pickled_optimiser = pickle.loads(pickle.dumps(optimiser))
        assert optimiser.paths.phase_path == pickled_optimiser.paths.phase_path



import configparser

from autofit import conf


def format_string_for_label(label: str) -> str:
    """
    Get the format for the label. Attempts to extract the key string associated with
    the dimension. Seems dodgy.

    Parameters
    ----------
    label
        A string label

    Returns
    -------
    format
        The format string (e.g {:.2f})
    """
    label_conf = conf.instance.label_format

    try:
        # noinspection PyProtectedMember
        for key, value in sorted(
            label_conf.parser._sections["format"].items(),
            key=lambda item: len(item[0]),
            reverse=True,
        ):
            if key in label:
                return value
    except KeyError:
        pass
    raise configparser.NoSectionError(
        "Could not find format for label {} in config at path {}".format(
            label, label_conf.path
        )
    )


def label_and_label_string(label0, label1, whitespace):
    return label0 + label1.rjust(whitespace - len(label0) + len(label1))


def label_and_value_string(label, value, whitespace, format_string=None):
    format_str = format_string or format_string_for_label(label)
    value = format_str.format(value)
    return label + value.rjust(whitespace - len(label) + len(value))


def label_value_and_limits_string(
    label, value, lower_limit, upper_limit, whitespace, format_string=None
):
    format_str = format_string or format_string_for_label(label)
    value = format_str.format(value)
    upper_limit = format_str.format(upper_limit)
    lower_limit = format_str.format(lower_limit)
    value = value + " (" + lower_limit + ", " + upper_limit + ")"
    return label + value.rjust(whitespace - len(label) + len(value))


def label_value_and_unit_string(label, value, unit, whitespace, format_string=None):
    format_str = format_string or format_string_for_label(label)
    value = (format_str + " {}").format(value, unit)
    return label + value.rjust(whitespace - len(label) + len(value))


def output_list_of_strings_to_file(file, list_of_strings):
    file = open(file, "w")
    file.write("".join(list_of_strings))
    file.close()


def within_radius_label_value_and_unit_string(
    prefix, radius, unit_length, value, unit_value, whitespace
):
    label = prefix + "_within_{:.2f}_{}".format(radius, unit_length)
    return label_value_and_unit_string(
        label=label, value=value, unit=unit_value, whitespace=whitespace
    )



from abc import ABC, abstractmethod

from autofit.tools.pipeline import ResultsCollection


class AbstractPromiseResult(ABC):
    def __init__(self, *result_path):
        self.result_path = result_path

    @property
    @abstractmethod
    def model(self):
        """
        A promise for an object in the model result. This might be a prior or prior model.
        """

    @property
    @abstractmethod
    def instance(self):
        """
        A promise for an object in the best fit result. This must be an instance or instance.
        """

    @abstractmethod
    def __getattr__(self, item):
        """
        Used to manage results paths
        """

    def __call__(self, *args, **kwargs):
        pass


class LastPromiseResult(AbstractPromiseResult):
    """
    A PromiseResult that does not require a phase. Refers to the latest Result in the collection with an object for
    the path specified in a promise.
    """

    @property
    def model(self):
        """
        A promise for an object in the model result. This might be a prior or prior model.
        """
        return LastPromise(result_path=self.result_path)

    @property
    def instance(self):
        """
        A promise for an object in the best fit result. This must be an instance or instance.
        """
        return LastPromise(result_path=self.result_path, is_instance=True)

    def __getattr__(self, item):
        return LastPromiseResult(*self.result_path, item)


class PromiseResult(AbstractPromiseResult):
    def __init__(self, phase, *result_path, assert_exists=True):
        """
        A wrapper for a phase that facilitates the generation of priors by consuming user defined paths.

        Parameters
        ----------
        phase
            A phase from which a promise is generated
        result_path
            The path through the results objects to the correct result. This is used for phase extensions when
            a result object can have child result objects
        assert_exists
            If this is true then an AttributeError is thrown if there is no object with the given path in the
            model of the origin phase
        """
        super().__init__(*result_path)
        self.phase = phase
        self.assert_exists = assert_exists

    @property
    def model(self):
        """
        A promise for an object in the model result. This might be a prior or prior model.
        """
        return Promise(
            self.phase, result_path=self.result_path, assert_exists=self.assert_exists
        )

    @property
    def instance(self):
        """
        A promise for an object in the best fit result. This must be an instance or instance.
        """
        return Promise(
            self.phase,
            result_path=self.result_path,
            is_instance=True,
            assert_exists=self.assert_exists,
        )

    def __getattr__(self, item):
        return PromiseResult(self.phase, *self.result_path, item, assert_exists=False)


class AbstractPromise(ABC):
    def __init__(self, *path, result_path, is_instance=False):
        """
        Place holder for an object in the object hierarchy. This is replaced at runtime by a prior, prior
        model, instance or instance

        Parameters
        ----------
        path
            The path to the promised object. e.g. if a phase has a model galaxies.lens.phi then the path
            will be ("galaxies", "lens", "phi")
        result_path
            The path through the result collection to the result object required. This is used for hyper phases
            where a result object can have child result objects for each phase extension.
        is_instance
            True if the promised object belongs to the instance result object
        """
        self.path = path
        self.is_instance = is_instance
        self.result_path = result_path

    def __call__(self, *args, **kwargs):
        pass

    def __getattr__(self, item):
        if item in ("phase", "path", "is_instance"):
            return super().__getattribute__(item)
        return Promise(
            *self.path, item, result_path=self.result_path, is_instance=self.is_instance
        )

    @abstractmethod
    def populate(self, results_collection):
        """
        Find the object that this promise corresponds to be getting results for a particular phase and then
        traversing those results using the path.

        Parameters
        ----------
        results_collection
            A collection of results from previous phases

        Returns
        -------
        obj
            The promised prior, prior model, instance or instance
        """

    def _populate_from_results(self, results):
        for item in self.result_path:
            results = getattr(results, item)
        model = results.instance if self.is_instance else results.model
        return model.object_for_path(self.path)


class Promise(AbstractPromise):
    def __init__(
        self, phase, *path, result_path, is_instance=False, assert_exists=True
    ):
        """
        Place holder for an object in the object hierarchy. This is replaced at runtime by a prior, prior
        model, instance or instance

        Parameters
        ----------
        phase
            The phase that holds or creates the promised object
        path
            The path to the promised object. e.g. if a phase has a model galaxies.lens.phi then the path
            will be ("galaxies", "lens", "phi")
        result_path
            The path through the result collection to the result object required. This is used for hyper phases
            where a result object can have child result objects for each phase extension.
        is_instance
            True if the promised object belongs to the instance result object
        assert_exists
            If this is true then an exception is raised if an object is not defined in the addressed phase's
            model. Hyper phases are a bit trickier so no assertion is made.
        """
        super().__init__(*path, result_path=result_path, is_instance=is_instance)
        self.phase = phase
        self.assert_exists = assert_exists
        if assert_exists:
            phase.model.object_for_path(path)

    def __getattr__(self, item):
        if item in ("phase", "path", "is_instance", "_populate_from_results"):
            return super().__getattribute__(item)
        return Promise(
            self.phase,
            *self.path,
            item,
            result_path=self.result_path,
            is_instance=self.is_instance,
            assert_exists=self.assert_exists,
        )

    def populate(self, results_collection):
        """
        Find the object that this promise corresponds to be getting results for a particular phase and then
        traversing those results using the path.

        Parameters
        ----------
        results_collection
            A collection of results from previous phases

        Returns
        -------
        obj
            The promised prior, prior model, instance or instance
        """
        results = results_collection.from_phase(self.phase.phase_name)
        return self._populate_from_results(results)


class LastPromise(AbstractPromise):
    """
    A promise that searches the results collection to find the latest result that contains an object with the
    specified path
    """

    def __getattr__(self, item):
        if item in ("phase", "path", "is_instance", "_populate_from_results"):
            return super().__getattribute__(item)
        return LastPromise(
            *self.path, item, result_path=self.result_path, is_instance=self.is_instance
        )

    def populate(self, results_collection: ResultsCollection):
        """
        Recover the instance or model associated with this promise from the latest result in the results collection
        where a matching path is found.
        
        Parameters
        ----------
        results_collection
            A collection of results to be searched in reverse

        Returns
        -------
        A prior, model or instance.
        
        Raises
        ------
        AttributeError
            If no matching prior is found
        """
        for results in results_collection.reversed:
            try:
                return self._populate_from_results(results)
            except AttributeError:
                pass
        raise AttributeError(
            f"No attribute found with path {self.path} in previous phase"
        )


last = LastPromiseResult()



import autofit as af


class TestPathDecorator:
    @staticmethod
    def assert_paths_as_expected(paths):
        assert paths.phase_name == "phase_name"
        assert paths.phase_tag == "phase_tag"
        assert paths.phase_folders == ["phase_folders"]

    def test_with_arguments(self):
        optimizer = af.NonLinearOptimizer(
            phase_name="phase_name",
            phase_tag="phase_tag",
            phase_folders=("phase_folders",),
        )
        self.assert_paths_as_expected(optimizer.paths)

    def test_positional(self):
        optimizer = af.NonLinearOptimizer("phase_name")
        paths = optimizer.paths

        assert paths.phase_name == "phase_name"

    def test_paths_argument(self):
        optimizer = af.NonLinearOptimizer(
            paths=af.Paths(
                phase_name="phase_name",
                phase_tag="phase_tag",
                phase_folders=("phase_folders",),
            )
        )
        self.assert_paths_as_expected(optimizer.paths)

    def test_combination_argument(self):
        optimizer = af.NonLinearOptimizer(
            "other",
            paths=af.Paths(
                phase_name="phase_name",
                phase_tag="phase_tag",
                phase_folders=("phase_folders",),
            ),
        )
        self.assert_paths_as_expected(optimizer.paths)
